//
//  複数オブジェクトの位置・向きをマウスで操作するためのクラス
//

/**
***  @file  ObjectLayout.h
***  @brief ObjectLayout クラスの定義
**/


#ifndef _OBJECT_LAYOUT_H_
#define _OBJECT_LAYOUT_H_


//#include <vecmath.h>
#include <Vector3.h>
#include <Point3.h>
#include <Point2.h>
#include <Matrix3.h>
#include <Matrix4.h>
#include <Color3.h>
#include <Color4.h>

#include <vector>
using namespace  std;


//
//  複数オブジェクトの位置・向きをマウスで操作するためのクラス
//
class  ObjectLayout
{
  protected:
	/*  内部用構造体・列挙型の定義  */
	
	// オブジェクト情報を表す構造体
	struct  Object
	{
		Point3f   pos;  // 位置
		Matrix3f  ori;  // 向き
		Vector3f  size; // サイズ（描画用）

		float     axis_length; // 軸の長さ
		Point2f   screen_pos;  // 画面上の座標
		Matrix4f  frame; // 位置・向き（キャッシュ用）
	};

	// 操作を表す列挙型
	enum  OperationEnum
	{
		OP_W_TRANSLATION,  // ワールド座標系での移動
		OP_M_TRANSLATION,
		OP_W_ROTATION,
		OP_M_ROTATION      // モデル座標系での回転
	};

	// 操作のためのハンドル（各軸・平面）を表す列挙型
	enum  HandleEnum
	{
		H_NONE = -1,
		H_MODE_CHANGE,

		H_M_X_AXIS,
		H_M_Y_AXIS,
		H_M_Z_AXIS,
		H_M_XY_PLANE,
		H_M_YZ_PLANE,
		H_M_ZX_PLANE,

		H_W_X_AXIS,
		H_W_Y_AXIS,
		H_W_Z_AXIS,
		H_W_XY_PLANE,
		H_W_YZ_PLANE,
		H_W_ZX_PLANE,

		NUM_HANDLE
	};

	// 操作に対応するマウス座標軸を表す列挙型
	enum  MouseControlEnum
	{
		NO_MOUSE_CONTROL,
		MOUSE_PLUS_X,
		MOUSE_MINUS_X,
		MOUSE_PLUS_Y,
		MOUSE_MINUS_Y
	};

	// 各種描画オプションを表す構造体
	struct  RenderOption
	{
		Color3f   box_color;
		Color3f   selected_box_color;

		float     box_width;
		float     axis_width;
		float     selected_axis_width;

		Color3f   x_axis_color;
		Color3f   y_axis_color;
		Color3f   z_axis_color;

		bool      change_axis_length;
		bool      enable_xray_mode;
		bool      enable_smooth;
	};

  protected:
	/*  動作設定情報  */

	// 描画オプション
	RenderOption  render_option;

	// デフォルトのオブジェクト情報
	Object  default_object;

  protected:
	/*  オブジェクト情報  */
	
	// 全オブジェクトの情報
	vector< Object >  objects;

  protected:
	/*  操作情報  */
	
	// 現在選択中のオブジェクト番号
	int  curr_object;

	// 現在の操作モード
	OperationEnum  curr_operation;

	// 現在有効なハンドル
	HandleEnum  active_handle;

	// 操作に対応するマウス座標軸
	MouseControlEnum  mouse_control;

	// ハンドル操作中かどうかのフラグ
	bool  on_control;

	// 操作ハンドルの画面上の座標
	Point2f  handle_pos[ NUM_HANDLE ];

	// 前回のマウス座標
	int  last_mouse_x;
	int  last_mouse_y;


  public:
	/*  初期化・終了処理  */
	
	// コンストラクタ
	ObjectLayout();

  public:
	/*  アクセサ  */
	
	size_t  GetNumObjects() const { return  objects.size(); }
	int  GetCurrentObject() const { return  curr_object; }
	Point3f &  GetPosition( int no ) { return  objects[ no ].pos; }
	Matrix3f &  GetOrientation( int no ) { return  objects[ no ].ori; }
	Matrix4f &  GetFrame( int no );
	Matrix4f &  GetTransposedFrame( int no );
	const Point3f &  GetPosition( int no ) const { return  objects[ no ].pos; }
	const Matrix3f &  GetOrientation( int no ) const { return  objects[ no ].ori; }
	RenderOption &  GetRenderOption() { return  render_option; }
	const char *  GetOperationMode();
	const float *  GetPositionArray( int no ) { return  & objects[ no ].pos.x; }
	const float *  GetOrientationArray( int no ) { return  & objects[ no ].ori.m00; }
	const float *  GetFrameArray( int no ) { GetFrame( no );  return  & objects[ no ].frame.m00; }
	const float *  GetTransposedFrameArray( int no ) { GetTransposedFrame( no );  return  & objects[ no ].frame.m00; }

  public:
	/*  オブジェクトの操作  */
	
	// オブジェクトの追加
	int  AddObject();

	// オブジェクトの削除
	int  DeleteObject();
	int  DeleteAllObjects();

	// オブジェクトの情報設定
	void  SetObjectPos( int no, const Point3f & p );
	void  SetObjectOri( int no, const Matrix3f & o );
	void  SetObjectFrame( int no, const Matrix4f & f );
	void  SetObjectSize( int no, const Vector3f & s );
	void  SetObjectSize( int no, float s );

  public:
	/*  イベントハンドラ  */

	// オブジェクト位置・視点の変更を通知
	void  Update();

	// マウス移動時の処理
	void  OnMoveMouse( int mx, int my );

	// マウスのボタンが押された時の処理
	void  OnMouseDown( int mx, int my );

	// マウスのボタンが離された時の処理
	void  OnMouseUp( int mx, int my );

	// オブジェクトの操作情報の描画
	void  Render();


  protected:
	/*  内部メソッド  */

	// 全オブジェクトをスクリーン座標系に投影
	void  ProjectObjects();

	// マウス位置に近いオブジェクトを探索
	int  FindObject( int mx, int my, int threshold );

	// マウス位置に近いハンドルを探索
	HandleEnum  FindHandle( int mx, int my, OperationEnum op, int threshold );
};



#endif // _OBJECT_LAYOUT_H_
