//
//  コンピュータグラフィックス特論Ⅱ
//  幾何形状モデル（Obj形式）の読み込み＆描画のサンプルプログラム
//


// 基本的なヘッダファイルのインクルード
#ifdef _WIN32
	#include <windows.h>
#endif
#include <stdio.h>
#include <GL/gl.h>

// 幾何形状データの定義のインクルード
#include "obj.h"


// バッファ長（サイズは適当）
#define  BUFFER_LENGTH      1024
#define  MAX_VECTOR_SIZE    4096
#define  MAX_TRIANGLE_SIZE  4096
#define  MAX_MTL_SIZE         32


//
//  Objファイルの読み込み
//
Obj *  LoadObj( const char * filename )
{
	FILE *  fp;
	char    line[ BUFFER_LENGTH ];
	char    name[ BUFFER_LENGTH ];
	int  i, j;
	Vector  vec;
	int  v_no[4], vt_no[4], vn_no[4];
	int  count;
	Mtl *  curr_mtl = NULL;

	// ファイルを開く
	fp = fopen( filename, "r" );
	if ( fp == NULL )
		return  NULL;

	// Obj構造体を初期化（ひとまず固定サイズの配列を割り当てる）
	Obj *  obj = new Obj();
	obj->num_vertices = 0;
	obj->num_normals = 0;
	obj->num_tex_coords = 0;
	obj->vertices = new Vector[ MAX_VECTOR_SIZE ];
	obj->normals = new Vector[ MAX_VECTOR_SIZE ];
	obj->tex_coords = new Vector[ MAX_VECTOR_SIZE ];
	obj->num_triangles = 0;
	obj->tri_v_no = new int[ MAX_TRIANGLE_SIZE * 3 ];
	obj->tri_vn_no = new int[ MAX_TRIANGLE_SIZE * 3 ];
	obj->tri_vt_no = new int[ MAX_TRIANGLE_SIZE * 3 ];
	obj->tri_material = new Mtl*[ MAX_TRIANGLE_SIZE * 3 ];
	obj->num_materials = 0;
	obj->materials = NULL;

	// ファイルから１行ずつ読み込み
	while ( fgets( line, BUFFER_LENGTH, fp ) != NULL )
	{
		// マテリアルの読み込み
		if ( strncmp( line, "mtllib", 6 ) == 0 )
		{
			// テキストを解析
			sscanf( line, "mtllib %s", name );

			// 指定されたファイル名のマテリアルデータを読み込み
			if ( strlen( name ) > 0 )
				LoadMtl( name, obj );
		}

		// マテリアルの変更
		if ( strncmp( line, "usemtl", 6 ) == 0 )
		{
			// テキストを解析
			sscanf( line, "usemtl %s", name );

			// 指定された名前のマテリアルデータを探索して記録
			for ( i=0; i<obj->num_materials; i++ )
			{
				if ( strcmp( name, obj->materials[ i ]->name ) == 0 )
				{
					curr_mtl = obj->materials[ i ];
					break;
				}
			}
		}

		// 頂点データの読み込み
		if ( line[0] == 'v' )
		{
			// 法線ベクトル（vn）
			if ( line[1] == 'n' )
			{
				// テキストを解析
				sscanf( line, "vn %f %f %f", &vec.x, &vec.y, &vec.z );

				// 法線ベクトル配列の末尾に格納
				if ( obj->num_normals < MAX_VECTOR_SIZE )
				{
					obj->normals[ obj->num_normals ] = vec;
					obj->num_normals ++;
				}
			}
			// テクスチャ座標（vt）
			else if ( line[1] == 't' )
			{
				// テキストを解析
				sscanf( line, "vt %f %f %f", &vec.x, &vec.y, &vec.z );

				// テクスチャ座標配列の末尾に格納
				if ( obj->num_tex_coords < MAX_VECTOR_SIZE )
				{
					obj->tex_coords[ obj->num_tex_coords ] = vec;
					obj->num_tex_coords ++;
				}
			}
			// 頂点座標（v）
			else
			{
				// テキストを解析
				sscanf( line, "v %f %f %f", &vec.x, &vec.y, &vec.z );

				// 法線ベクトル配列の末尾に格納
				if ( obj->num_vertices < MAX_VECTOR_SIZE )
				{
					obj->vertices[ obj->num_vertices ] = vec;
					obj->num_vertices ++;
				}
			}
		}

		// ポリゴンデータの読み込み
		if ( line[0] == 'f' )
		{
			// 未実装（各自実装）

			// テキストを解析（三角形・テクスチャ座標なしの場合）
			count = sscanf( line, "f %i//%i %i//%i %i//%i", &v_no[0], &vn_no[0], &v_no[1], &vn_no[1], &v_no[2], &vn_no[2] );

			// 解析に成功したらポリゴンデータを記録
			if ( count == 6 )
			{
				i = obj->num_triangles * 3;
				for ( j=0; j<3; j++ )
				{
					obj->tri_v_no[ i+j ] = v_no[ j ] - 1;  // Obj形式ではインデックス番号は１から始まるので、－１して０から始まるようにする
					obj->tri_vn_no[ i+j ] = vn_no[ j ] - 1;
					obj->tri_vt_no[ i+j ] = vt_no[ j ] - 1;
				}
				obj->tri_material[ obj->num_triangles ] = curr_mtl;
				obj->num_triangles ++;
			}
			// 解析に失敗したら別のフォーマットを試す
			{
				// ......

				// 未実装（各自実装）
			}

			// ポリゴン数が確保した配列の大きさを超えたら強制終了
			if ( obj->num_triangles >= MAX_TRIANGLE_SIZE )
				break;
		}
	};

	// 必要な配列を確保しなおす（面倒なのでとりあえず頂点座標配列のみ、後は各自追加）
	Vector *  new_array;
	new_array = new Vector[ obj->num_vertices ];
	memcpy( new_array, obj->vertices, sizeof( Vector ) * obj->num_vertices );
	delete[]  obj->vertices;
	obj->vertices = new_array;

	// ファイルを閉じる
	fclose( fp );

	// 読み込んだオブジェクトデータを返す
	return  obj;
}


//
//  Mtlファイルの読み込み
//
void  LoadMtl( const char * filename, Obj * obj )
{
	FILE *  fp;
	char    line[ BUFFER_LENGTH ];
	char    name[ BUFFER_LENGTH ];
	Color   color;
	Mtl *   curr_mtl = NULL;

	// ファイルを開く
	fp = fopen( filename, "r" );
	if ( fp == NULL )
		return;

	// Mtl配列を初期化（ひとまず固定サイズの配列を割り当てる）
	obj->num_materials = 0;
	obj->materials = new Mtl*[ MAX_MTL_SIZE ];

	// ファイルから１行ずつ読み込み
	while ( fgets( line, BUFFER_LENGTH, fp ) != NULL )
	{
		// マテリアルデータの追加
		if ( strncmp( line, "newmtl", 6 ) == 0 )
		{
			// テキストを解析
			sscanf( line, "newmtl %s", name );
			if ( strlen( name ) == 0 ) 
				continue;

			// マテリアルデータの作成
			curr_mtl = new Mtl();
			curr_mtl->name = new char[ strlen( name ) + 1 ];
			strcpy( curr_mtl->name, name );
			curr_mtl->kd.r = 0.8f;
			curr_mtl->kd.g = 0.8f;
			curr_mtl->kd.b = 0.8f;

			// マテリアルデータを配列に記録
			obj->materials[ obj->num_materials ] = curr_mtl;
			obj->num_materials ++;
		}

		// 反射特性データの読み込み
		if ( line[0] == 'K' )
		{
			// 拡散反射特性（Kd）
			if ( line[1] == 'd' )
			{
				// テキストを解析
				sscanf( line, "Kd %f %f %f", &color.r, &color.g, &color.b );

				// 拡散反射特性（Kd）を記録
				if ( curr_mtl )
					curr_mtl->kd = color;
			}
		}

	}

	// ファイルを閉じる
	fclose( fp );
}


//
//  幾何形状モデルのスケーリング（スケーリング後のサイズを返す）
//
void  ScaleObj( Obj * obj, float max_size, float * size_x, float * size_y, float * size_z )
{
	if ( !obj || ( obj->num_vertices == 0 ) )
		return;

	// サイズ計算
	Vector  min, max;
	min = max = obj->vertices[ 0 ];
	int  i;
	for ( i=0; i<obj->num_vertices; i++ )
	{
		const Vector &  p = obj->vertices[ i ];
		if ( p.x < min.x )  min.x = p.x;
		if ( p.y < min.y )  min.y = p.y;
		if ( p.z < min.z )  min.z = p.z;
		if ( p.x > max.x )  max.x = p.x;
		if ( p.y > max.y )  max.y = p.y;
		if ( p.z > max.z )  max.z = p.z;
	}

	// スケーリング
	float  size, scale;
	size = ( ( max.x - min.x ) > ( max.z - min.z ) ) ? ( max.x - min.x ) : ( max.z - min.z );
	scale = max_size / size;
	for ( i=0; i<obj->num_vertices; i++ )
	{
		obj->vertices[ i ].x *= scale;
		obj->vertices[ i ].y *= scale;
		obj->vertices[ i ].z *= scale;
	}

	// スケーリング後の大きさを返す
	if ( size_x )
		*size_x = ( max.x - min.x ) * scale;
	if ( size_y )
		*size_y = ( max.y - min.y ) * scale;
	if ( size_z )
		*size_z = ( max.z - min.z ) * scale;
}


//
//  幾何形状モデル（Obj形状）の描画
//
void  RenderObj( Obj * obj )
{
	int  i, j, no;
	Mtl *  curr_mtl = NULL;

	glBegin( GL_TRIANGLES );
	for ( i=0; i<obj->num_triangles; i++ )
	{
		// マテリアルの切り替え
		if ( ( obj->num_materials > 0 ) && ( obj->tri_material[ i ] != curr_mtl ) )
		{
			curr_mtl = obj->tri_material[ i ];
			glColor3f( curr_mtl->kd.r, curr_mtl->kd.g, curr_mtl->kd.b );
		}

		// 三角面の各頂点データの指定
		for ( j=0; j<3; j++ )
		{
			// 法線ベクトル
			no = obj->tri_vn_no[ i*3 + j ];
			const Vector &  vn = obj->normals[ no ];
			glNormal3f( vn.x, vn.y, vn.z );

			// 頂点座標
			no = obj->tri_v_no[ i*3 + j ];
			const Vector &  v = obj->vertices[ no ];
			glVertex3f( v.x, v.y, v.z );
		}
	}
	glEnd();
}


//
//  幾何形状モデル（Obj形状）の描画（固定色で描画）
//
void  RenderObjUnicolor( const Obj * obj, float color_r, float color_g, float color_b, float color_a )
{
	// ※レポート課題
#define  REPORT
#ifdef  REPORT

	int  i, j, no;
	Mtl *  curr_mtl = NULL;

	glBegin( GL_TRIANGLES );
	glColor4f( color_r, color_g, color_b, color_a );
//	glNormal3f( 0.0f, 1.0f, 0.0f );
	for ( i=0; i<obj->num_triangles; i++ )
	{
/*		// マテリアルの切り替え
		if ( ( obj->num_materials > 0 ) && ( obj->tri_material[ i ] != curr_mtl ) )
		{
			curr_mtl = obj->tri_material[ i ];
			glColor3f( curr_mtl->kd.r, curr_mtl->kd.g, curr_mtl->kd.b );
		}
*/
		// 三角面の各頂点データの指定
		for ( j=0; j<3; j++ )
		{
			// 法線ベクトル
			no = obj->tri_vn_no[ i*3 + j ];
			const Vector &  vn = obj->normals[ no ];
			glNormal3f( vn.x, vn.y, vn.z );

			// 頂点座標
			no = obj->tri_v_no[ i*3 + j ];
			const Vector &  v = obj->vertices[ no ];
			glVertex3f( v.x, v.y, v.z );
		}
	}
	glEnd();

#endif // REPORT
}


// 影の描画の設定（ブレンディングやステンシルバッファの使用の有無の切り替え）
static bool  shadow_blend_off = false;
static bool  shadow_stencil_off = false;


//
//  ポリゴン投影による影の描画
//
void  RenderObjShadow( const Obj * obj, const float obj_matrix[ 16 ], float light_dir_x, float light_dir_y, float light_dir_z, float color_r, float color_g, float color_b, float color_a )
{
	// ※レポート課題
#ifdef  REPORT
	// 現在の描画設定を取得（描画終了後に元の設定に戻すため）
	GLboolean  b_cull_face, b_blend, b_lighting, b_stencil;
	glGetBooleanv( GL_CULL_FACE, &b_cull_face );
	glGetBooleanv( GL_BLEND, &b_blend );
	glGetBooleanv( GL_LIGHTING, &b_lighting );
	glGetBooleanv( GL_STENCIL_TEST, &b_stencil );

	// 描画設定の変更
	if ( b_lighting )
		glDisable( GL_LIGHTING );
	if ( !b_cull_face )
		glEnable( GL_CULL_FACE );
	if ( !b_blend )
		glEnable( GL_BLEND );

	// ブレンディングの設定
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	// ステンシルバッファの設定

	// ※レポート課題
#ifdef  REPORT
	glEnable( GL_STENCIL_TEST );
	glStencilFunc( GL_NOTEQUAL, 1, 1 );
	glStencilOp( GL_KEEP, GL_KEEP, GL_REPLACE );
#endif // REPORT

	// 動作確認のための描画設定の変更
	if ( shadow_blend_off )
		glDisable( GL_BLEND );
	if ( shadow_stencil_off )
		glDisable( GL_STENCIL_TEST );


	// 現在の変換行列を一時保存
	glPushMatrix();

	// ポリゴンモデルを地面に投影して描画するための変換行列を設定
	//（この時点で、ワールド座標系からカメラ座標系への変換行列が設定されているものとする）


	// ※レポート課題
#ifdef  REPORT

	// 地面への投影行列を計算
	float  mat[ 16 ];
	mat[ 0 ] = 1.0f;  mat[ 4 ] = - light_dir_x / light_dir_y;  mat[ 8 ] = 0.0f;  mat[ 12 ] = 0.0f;
	mat[ 1 ] = 0.0f;  mat[ 5 ] = 0.0f;                         mat[ 9 ] = 0.0f;  mat[ 13 ] = 0.01f;
	mat[ 2 ] = 0.0f;  mat[ 6 ] = - light_dir_z / light_dir_y;  mat[ 10 ] = 1.0f;  mat[ 14 ] = 0.0f;
	mat[ 3 ] = 0.0f;  mat[ 7 ] = 0.0f;                         mat[ 11 ] = 0.0f;  mat[ 15 ] = 1.0f;

	// 地面への投影行列をかける
	glMultMatrixf( mat );

	// モデル座標からワールド座標への変換行列をかける
	glMultMatrixf( obj_matrix );

#endif // REPORT


	// 影の描画、幾何形状モデルを指定色で描画
	RenderObjUnicolor( obj, color_r, color_g, color_b, color_a );

	// 一時保存しておいた変換行列を復元
	glPopMatrix();

	// 描画設定を復元
	if ( b_lighting )
		glEnable( GL_LIGHTING );
	if ( b_cull_face )
		glEnable( GL_CULL_FACE );
	if ( !b_blend )
		glDisable( GL_BLEND );
	if ( !b_stencil )
		glDisable( GL_STENCIL_TEST );
#endif // REPORT
}


//
//  幾何形状モデルを頂点配列を使って描画可能なモデルに変換
//
void  ConvertObjForVertexArrays( Obj * obj )
{
	// 各自実装
}


//
//  幾何形状モデルを頂点配列を使って描画
//
void  RenderObjWithVertexArrays( Obj * obj )
{
	// 各自実装
}


